import numpy as np
import pandas as pd

from allensdk.brain_observatory.behavior.behavior_project_cache.\
    project_apis.data_io import VisualBehaviorNeuropixelsProjectCloudApi
from allensdk.brain_observatory.behavior.behavior_project_cache.\
    project_cache_base import ProjectCacheBase
from allensdk.brain_observatory.behavior.behavior_session import \
    BehaviorSession
from allensdk.brain_observatory.ecephys.behavior_ecephys_session import \
    BehaviorEcephysSession


class VisualBehaviorNeuropixelsProjectCache(ProjectCacheBase):
    """ Entrypoint for accessing Visual Behavior Neuropixels data.

    Supports access to metadata tables:
    get_ecephys_session_table()
    get_behavior_session_table()
    get_probe_table()
    get_channel_table()
    get_unit_table

    Provides methods for instantiating session objects
    from the nwb files:
    get_ecephys_session() to load BehaviorEcephysSession
    get_behavior_sesion() to load BehaviorSession

    Provides tools for downloading data:

    Will download data from the s3 bucket if session nwb file is not
    in the local cache, otherwise will use file from the cache.
    """

    PROJECT_NAME = "visual-behavior-neuropixels"
    BUCKET_NAME = "visual-behavior-neuropixels-data"

    def __init__(
            self,
            fetch_api: VisualBehaviorNeuropixelsProjectCloudApi,
            fetch_tries: int = 2
    ):
        super().__init__(fetch_api=fetch_api, fetch_tries=fetch_tries)

    @classmethod
    def cloud_api_class(cls):
        return VisualBehaviorNeuropixelsProjectCloudApi

    def get_ecephys_session_table(
            self,
            filter_abnormalities: bool = True
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        filter_abnormalities: bool
            If True, do not return rows corresponding to
            sessions with identified abnormalities in
            histology or activity

        Returns
        -------
        ecephys_sessions_table: pd.DataFrame
            pandas dataframe representing metadata for all
            ecephys sessions in the data release
        """

        sessions_table = self.fetch_api.get_ecephys_session_table()

        if filter_abnormalities:
            sessions_table = sessions_table.loc[
                    np.logical_and(
                        sessions_table.abnormal_histology.isna(),
                        sessions_table.abnormal_activity.isna())]

        return sessions_table

    def get_behavior_session_table(self) -> pd.DataFrame:
        """
        Returns
        -------
        behavior_sessions_table: pd.DataFrame
            pandas dataframe representing metadata for all
            behavior sessions in the data release
        """
        return self.fetch_api.get_behavior_session_table()

    def get_probe_table(self) -> pd.DataFrame:
        """
        Returns
        -------
        probes table: pd.DataFrame

        Columns:
            - id: probe id
            - name: probe name
            - location: probe location
            - lfp_sampling_rate: LFP sampling rate
            - has_lfp_data: Whether this probe has LFP data
        """
        return self.fetch_api.get_probe_table()

    def get_channel_table(self) -> pd.DataFrame:
        """
        Returns
        -------
        channels table: pd.DataFrame

        Index: id
        Columns:
            - properties of `allensdk.ecephys._channel.Channel`
            except for 'impedance'
        """
        return self.fetch_api.get_channel_table()

    def get_unit_table(self) -> pd.DataFrame:
        """
        Returns
        -------
        units table: pd.DataFrame

        Columns:
            - properties of `allensdk.ecephys._unit.Unit`
            except for 'spike_times', 'spike_amplitudes', 'mean_waveforms'
            which are returned separately
        """
        return self.fetch_api.get_unit_table()

    def get_ecephys_session(
            self,
            ecephys_session_id: int
    ) -> BehaviorEcephysSession:
        """
        Loads all data for `ecephys_session_id` into an
        `allensdk.ecephys.behavior_ecephys_session.BehaviorEcephysSession`
        instance

        Parameters
        ----------
        ecephys_session_id: int
            The ecephys session id

        Returns
        -------
        `allensdk.ecephys.behavior_ecephys_session.BehaviorEcephysSession`
        instance

        """
        return self.fetch_api.get_ecephys_session(ecephys_session_id)

    def get_behavior_session(
            self,
            behavior_session_id: int
    ) -> BehaviorSession:
        """
        Loads all data for `behavior_session_id` into an
        `allensdk.brain_observatory.behavior.behavior_session.BehaviorSession`
        instance

        Parameters
        ----------
        behavior_session_id: int
            The behavior session id

        Returns
        -------
        `allensdk.brain_observatory.behavior.behavior_session.BehaviorSession`
        instance

        """
        return self.fetch_api.get_behavior_session(behavior_session_id)
