import pathlib
from typing import Union

import numpy as np
from allensdk.api.cloud_cache.cloud_cache import S3CloudCache
from allensdk.api.cloud_cache.file_attributes import CacheFileAttributes
from allensdk.brain_observatory.behavior.data_objects.stimuli.stimulus_templates import (  # noqa: E501
    StimulusMovieTemplateFactory,
)


class NaturalMovieOneCache(S3CloudCache):
    def __init__(self, cache_dir: Union[str, pathlib.Path], bucket_name: str):
        super().__init__(
            cache_dir=cache_dir,
            bucket_name=bucket_name,
            project_name=None,
            ui_class_name=None,
        )

        # Set the file attributes by hand. This is used to get around needing
        # to run the data release tool and create/download a manifest file.
        # The hash has been pre-calculated from the file_hash_from_path
        # method in allensdk/api/cloud_cache/utils.py
        self._file_attributes = CacheFileAttributes(
            url="https://visual-behavior-ophys-data.s3.us-west-2.amazonaws.com/visual-behavior-ophys/resources/Movie_TOE1.npy",  # noqaL E501
            version_id="Czht4ouDiTvs4CC_s8hsaE7VtVt_E9rO",
            file_hash="7e44cba154b29e1511ab8e5453b7aa5070f1ae456724b5b2541c97c052fbd80aebf159e5f933ab319bda8fdab7b863a096cdb44f129abd20a8c4cc791af4bc41",  # noqa E501
            local_path=pathlib.Path(cache_dir) / "Movie_TOE1.npy",
        )

    def _list_all_manifests(self) -> list:
        """
        Return a list of all of the file names of the manifests associated
        with this dataset
        """
        return None

    def get_file_attributes(self, file_id):
        """
        Retrieve file attributes for a given file_id from the meatadata.

        Parameters
        ----------
        file_id: str or int
            The unique identifier of the file to be accessed (not used in this
            overwrite of the method)

        Returns
        -------
        CacheFileAttributes
        """
        return self._file_attributes

    def get_raw_movie(self):
        """Download the raw movie data from the cloud and return it as a numpy
        array.

        Returns
        -------
        raw_movie : np.ndarray
        """
        return np.load(self.download_data(None))

    def get_processed_template_movie(self, n_workers=None):
        """Download the movie if needed and process it into warped and unwarped
        frames as presented to the mouse. The DataFrame is indexed with the
        same frame index as shown in the stimulus presentation table.

        The processing of the movie requires signicant processing and its
        return size is very large so take care in requesting this data.

        Parameters
        ----------
        n_workers : int
            Number of processes to use to transform the movie to what was shown
            on the monitor. Default=None (use all cores).

        Returns
        -------
        processed_movie : pd.DataFrame
        """
        movie_data = self.get_raw_movie()
        movie_template = StimulusMovieTemplateFactory.from_unprocessed(
            movie_name="natural_movie_one",
            movie_frames=movie_data,
            n_workers=n_workers,
        )
        return movie_template.to_dataframe(
            index_name="movie_frame_index", index_type="int"
        )
