from typing import Union, Optional
from pathlib import Path
import logging

from allensdk.api.cloud_cache.cloud_cache import (
    S3CloudCache, LocalCache, StaticLocalCache)

from allensdk.brain_observatory.behavior.behavior_project_cache \
    .utils import version_check


class ProjectCloudApiBase(object):
    """API for downloading data released on S3 and returning tables.

    Parameters
    ----------
    cache: S3CloudCache
        an instantiated S3CloudCache object, which has already run
        `self.load_manifest()` which populates the columns:
          - metadata_file_names
          - file_id_column
    skip_version_check: bool
        whether to skip the version checking of pipeline SDK version
        vs. running SDK version, which may raise Exceptions. (default=False)
    local: bool
        Whether to operate in local mode, where no data will be downloaded
        and instead will be loaded from local
    """
    def __init__(
        self,
        cache: Union[S3CloudCache, LocalCache, StaticLocalCache],
        skip_version_check: bool = False,
        local: bool = False
    ):

        self.cache = cache
        self.skip_version_check = skip_version_check
        self._local = local
        self.load_manifest()

    def load_manifest(self, manifest_name: Optional[str] = None):
        """
        Load the specified manifest file into the CloudCache

        Parameters
        ----------
        manifest_name: Optional[str]
            Name of manifest file to load. If None, load latest
            (default: None)
        """
        if manifest_name is None:
            self.cache.load_last_manifest()
        else:
            self.cache.load_manifest(manifest_name)

        if self.cache._manifest.metadata_file_names is None:
            raise RuntimeError(f"{type(self.cache)} object has no metadata "
                               f"file names. Check contents of the loaded "
                               f"manifest file: {self.cache._manifest_name}")

        if not self.skip_version_check:
            data_sdk_version = [i for i in self.cache._manifest._data_pipeline
                                if i['name'] == "AllenSDK"][0]["version"]
            version_check(
                self.cache._manifest.version,
                data_sdk_version,
                cmin=self.MANIFEST_COMPATIBILITY[0],
                cmax=self.MANIFEST_COMPATIBILITY[1])

        self.logger = logging.getLogger(self.__class__.__name__)

        self._load_manifest_tables()

    def _load_manifest_tables(self):

        raise NotImplementedError

    @classmethod
    def from_s3_cache(cls, cache_dir: Union[str, Path],
                      bucket_name: str,
                      project_name: str,
                      ui_class_name: str) -> "ProjectCloudApiBase":
        """instantiates this object with a connection to an s3 bucket and/or
        a local cache related to that bucket.

        Parameters
        ----------
        cache_dir: str or pathlib.Path
            Path to the directory where data will be stored on the local system

        bucket_name: str
            for example, if bucket URI is 's3://mybucket' this value should be
            'mybucket'

        project_name: str
            the name of the project this cache is supposed to access. This
            project name is the first part of the prefix of the release data
            objects. I.e. s3://<bucket_name>/<project_name>/<object tree>

        ui_class_name: str
            Name of user interface class (used to populate error messages)

        Returns
        -------
        BehaviorProjectCloudApi instance

        """
        cache = S3CloudCache(cache_dir,
                             bucket_name,
                             project_name,
                             ui_class_name=ui_class_name)
        return cls(cache)

    @classmethod
    def from_local_cache(
        cls,
        cache_dir: Union[str, Path],
        project_name: str,
        ui_class_name: str,
        use_static_cache: bool = False
    ) -> "ProjectCloudApiBase":
        """instantiates this object with a local cache.

        Parameters
        ----------
        cache_dir: str or pathlib.Path
            Path to the directory where data will be stored on the local system

        project_name: str
            the name of the project this cache is supposed to access. This
            project name is the first part of the prefix of the release data
            objects. I.e. s3://<bucket_name>/<project_name>/<object tree>

        ui_class_name: str
            Name of user interface class (used to populate error messages)

        Returns
        -------
        ProjectCloudApiBase instance

        """
        if use_static_cache:
            cache = StaticLocalCache(
                cache_dir,
                project_name,
                ui_class_name=ui_class_name
            )
        else:
            cache = LocalCache(
                cache_dir,
                project_name,
                ui_class_name=ui_class_name
            )
        return cls(cache, local=True)

    def _get_metadata_path(self, fname: str):
        if self._local:
            path = self._get_local_path(fname=fname)
        else:
            path = self.cache.download_metadata(fname=fname)
        return path

    def _get_data_path(self, file_id: str):
        if self._local:
            data_path = self._get_local_path(file_id=file_id)
        else:
            data_path = self.cache.download_data(file_id=file_id)
        return data_path

    def _get_local_path(self, fname: Optional[str] = None, file_id:
                        Optional[str] = None):
        if fname is None and file_id is None:
            raise ValueError('Must pass either fname or file_id')

        if fname is not None and file_id is not None:
            raise ValueError('Must pass only one of fname or file_id')

        if fname is not None:
            path = self.cache.metadata_path(fname=fname)
        else:
            path = self.cache.data_path(file_id=file_id)

        exists = path['exists']
        local_path = path['local_path']
        if not exists:
            raise FileNotFoundError(f'You started a cache without a '
                                    f'connection to s3 and {local_path} is '
                                    'not already on your system')
        return local_path
