import json
import numpy as np
from typing import Union, Tuple
from pathlib import Path

import h5py
from cachetools import cached, LRUCache
from cachetools.keys import hashkey

import pandas as pd

from allensdk.internal.api import PostgresQueryMixin
from allensdk.internal.core.lims_utilities import safe_system_path
from allensdk.internal.core import DataFile


def from_json_cache_key(cls, dict_repr: dict):
    return hashkey(json.dumps(dict_repr))


def from_lims_cache_key(cls, db, ophys_experiment_id: int):
    return hashkey(ophys_experiment_id)


class EventDetectionFile(DataFile):
    """A DataFile which contains methods for accessing and loading
    events.
    """

    def __init__(self, filepath: Union[str, Path]):
        super().__init__(filepath=filepath)

    @classmethod
    @cached(cache=LRUCache(maxsize=10), key=from_json_cache_key)
    def from_json(cls, dict_repr: dict) -> "EventDetectionFile":
        filepath = dict_repr["events_file"]
        return cls(filepath=filepath)

    @classmethod
    @cached(cache=LRUCache(maxsize=10), key=from_lims_cache_key)
    def from_lims(
        cls, db: PostgresQueryMixin,
        ophys_experiment_id: Union[int, str]
    ) -> "EventDetectionFile":
        query = f'''
            SELECT wkf.storage_directory || wkf.filename AS event_detection_filepath
            FROM ophys_experiments oe
            LEFT JOIN well_known_files wkf ON wkf.attachable_id = oe.id
            JOIN well_known_file_types wkft ON wkf.well_known_file_type_id = wkft.id
            WHERE wkft.name = 'OphysEventTraceFile'
                AND oe.id = {ophys_experiment_id};
        '''  # noqa E501
        filepath = safe_system_path(db.fetchone(query, strict=True))
        return cls(filepath=filepath)

    @staticmethod
    def load_data(filepath: Union[str, Path]) -> \
            Tuple[np.ndarray, pd.DataFrame]:
        with h5py.File(filepath, 'r') as f:
            events = f['events'][:]
            lambdas = f['lambdas'][:]
            noise_stds = f['noise_stds'][:]
            roi_ids = f['roi_names'][:]

        df = pd.DataFrame({
            'lambda': lambdas,
            'noise_std': noise_stds,
            'cell_roi_id': roi_ids
        })
        return events, df
