from typing import Optional
import numpy as np
import pandas as pd
from hdmf.backends.hdf5 import H5DataIO
from pynwb import NWBFile

from allensdk.brain_observatory.behavior.data_files.event_detection_file \
    import \
    EventDetectionFile
from allensdk.core import DataObject
from allensdk.core import \
    DataFileReadableInterface, NwbReadableInterface
from allensdk.core import \
    NwbWritableInterface
from allensdk.brain_observatory.behavior.data_objects.cell_specimens\
    .rois_mixin import \
    RoisMixin
from allensdk.brain_observatory.behavior.event_detection import \
    filter_events_array
from allensdk.brain_observatory.behavior.write_nwb.extensions\
    .event_detection.ndx_ophys_events import \
    OphysEventDetection


class Events(DataObject, RoisMixin, DataFileReadableInterface,
             NwbReadableInterface, NwbWritableInterface):
    """Events
    columns:
        events: np.array
        lambda: float
        noise_std: float
        cell_roi_id: int
    """
    def __init__(self,
                 events: np.ndarray,
                 events_meta: pd.DataFrame,
                 frame_rate_hz: float,
                 filter_scale_seconds: float = 2.0/31.0,
                 filter_n_time_steps: int = 20):
        """
        Parameters
        ----------
        events
            events
        events_meta
            lambda, noise_std, cell_roi_id for each roi
        frame_rate_hz
            The frame rate of the experiment in Hz;
            unitless filter_scale for filter_events_array
            is filter_scale_seconds*frame_rate_hz
        filter_scale_seconds
            The filter scale for the events in seconds
        filter_n_time_steps
            See filter_events_array for description
        """

        filtered_events = filter_events_array(
            arr=events,
            scale=filter_scale_seconds*frame_rate_hz,
            n_time_steps=filter_n_time_steps)

        # Convert matrix to list of 1d arrays so that it can be stored
        # in a single column of the dataframe
        events = [x for x in events]
        filtered_events = [x for x in filtered_events]

        df = pd.DataFrame({
            'events': events,
            'filtered_events': filtered_events,
            'lambda': events_meta['lambda'],
            'noise_std': events_meta['noise_std'],
            'cell_roi_id': events_meta['cell_roi_id']
        })
        super().__init__(name='events', value=df)

    @classmethod
    def from_data_file(cls,
                       events_file: EventDetectionFile,
                       filter_scale_seconds: float = 2.0/31.0,
                       filter_n_time_steps: int = 20,
                       frame_rate_hz: Optional[float] = None) -> "Events":
        events, events_meta = events_file.data
        return cls(events=events,
                   events_meta=events_meta,
                   filter_scale_seconds=filter_scale_seconds,
                   filter_n_time_steps=filter_n_time_steps,
                   frame_rate_hz=frame_rate_hz)

    @classmethod
    def from_nwb(cls,
                 nwbfile: NWBFile,
                 filter_scale_seconds: float = 2.0/31.0,
                 filter_n_time_steps: int = 20,
                 frame_rate_hz: Optional[float] = None) -> "Events":
        event_detection = nwbfile.processing['ophys']['event_detection']
        # NOTE: The rois with events are stored in event detection
        partial_cell_specimen_table = event_detection.rois.to_dataframe()

        events = event_detection.data[:]

        # events stored time x roi. Change back to roi x time
        events = events.T

        events_meta = pd.DataFrame({
            'cell_roi_id': partial_cell_specimen_table.index,
            'lambda': event_detection.lambdas[:],
            'noise_std': event_detection.noise_stds[:]
        })
        return cls(events=events,
                   events_meta=events_meta,
                   filter_scale_seconds=filter_scale_seconds,
                   filter_n_time_steps=filter_n_time_steps,
                   frame_rate_hz=frame_rate_hz)

    def to_nwb(self, nwbfile: NWBFile) -> NWBFile:
        events = self.value.set_index('cell_roi_id')

        ophys_module = nwbfile.processing['ophys']
        dff_interface = ophys_module.data_interfaces['dff']
        traces = dff_interface.roi_response_series['traces']
        seg_interface = ophys_module.data_interfaces['image_segmentation']

        cell_specimen_table = (
            seg_interface.plane_segmentations['cell_specimen_table'])
        cell_specimen_df = cell_specimen_table.to_dataframe()

        # We only want to store the subset of rois that have events data
        rois_with_events_indices = [cell_specimen_df.index.get_loc(label)
                                    for label in events.index]
        roi_table_region = cell_specimen_table.create_roi_table_region(
            description="Cells with detected events",
            region=rois_with_events_indices)

        events_data = np.vstack(events['events'])
        events = OphysEventDetection(
            # time x rois instead of rois x time
            # store using compression since sparse
            data=H5DataIO(events_data.T, compression=True),

            lambdas=events['lambda'].values,
            noise_stds=events['noise_std'].values,
            unit='N/A',
            rois=roi_table_region,
            timestamps=traces.timestamps
        )

        ophys_module.add_data_interface(events)

        return nwbfile
