import numpy as np
import pandas as pd
from allensdk.brain_observatory.behavior.data_files.neuropil_corrected_file import (  # noqa: E501
    NeuropilCorrectedFile,
)
from allensdk.brain_observatory.behavior.data_objects.cell_specimens.rois_mixin import (  # noqa: E501
    RoisMixin,
)
from allensdk.core import (
    DataFileReadableInterface,
    DataObject,
    NwbReadableInterface,
    NwbWritableInterface,
)
from pynwb import NWBFile
from pynwb.ophys import Fluorescence


class CorrectedFluorescenceTraces(
    DataObject,
    RoisMixin,
    DataFileReadableInterface,
    NwbReadableInterface,
    NwbWritableInterface,
):
    """A data container to load, access, and store the
    corrected_fluorescence_traces dataframe. Corrected fluorescence traces
    are neuropil corrected and demixed.
    """

    def __init__(self, traces: pd.DataFrame):
        """

        Parameters
        ----------
        traces
            index cell_roi_id
            columns:
                corrected_fluorescence: (list of float)
                    fluorescence values (arbitrary units)
                RMSE: (float)
                    error values (arbitrary units)
                r:
                    r values (arbitrary units)
        """
        super().__init__(name="corrected_fluorescence_traces", value=traces)

    @classmethod
    def from_nwb(cls, nwbfile: NWBFile) -> "CorrectedFluorescenceTraces":
        corr_fluorescence_traces_nwb = nwbfile.processing[
            "ophys"
        ].data_interfaces["corrected_fluorescence"]
        # f traces stored as timepoints x rois in NWB
        # We want rois x timepoints, hence the transpose
        f_traces = (
            corr_fluorescence_traces_nwb.roi_response_series["traces"]
            .data[:]
            .T.copy()
        )
        roi_ids = (
            corr_fluorescence_traces_nwb.roi_response_series["traces"]
            .rois.table.id[:]
            .copy()
        )
        # TODO: Remove try/except once VBO released.
        try:
            r_values = (
                corr_fluorescence_traces_nwb.roi_response_series["r"]
                .data[:]
                .copy()
            )
            rmse = (
                corr_fluorescence_traces_nwb.roi_response_series["RMSE"]
                .data[:]
                .copy()
            )
            data_dict = {
                "corrected_fluorescence": [x for x in f_traces],
                "r": r_values,
                "RMSE": rmse,
            }
        except KeyError:
            data_dict = {"corrected_fluorescence": [x for x in f_traces]}
        df = pd.DataFrame(
            data=data_dict,
            index=pd.Index(data=roi_ids, name="cell_roi_id"),
        )
        return CorrectedFluorescenceTraces(traces=df)

    @classmethod
    def from_data_file(
        cls, neuropil_corrected_file: NeuropilCorrectedFile
    ) -> "CorrectedFluorescenceTraces":
        corrected_fluorescence_traces = neuropil_corrected_file.data
        return cls(traces=corrected_fluorescence_traces)

    def to_nwb(self, nwbfile: NWBFile) -> NWBFile:
        corrected_fluorescence_traces = self.value["corrected_fluorescence"]
        rmse = self.value["RMSE"].values
        r_values = self.value["r"].values
        # Convert from Series of lists to numpy array
        # of shape ROIs x timepoints
        traces = np.stack([x for x in corrected_fluorescence_traces])

        # Create/Add corrected_fluorescence_traces modules and interfaces:
        ophys_module = nwbfile.processing["ophys"]

        roi_table_region = (
            nwbfile.processing["ophys"]
            .data_interfaces["dff"]
            .roi_response_series["traces"]
            .rois
        )  # noqa: E501
        ophys_timestamps = (
            ophys_module.get_data_interface("dff")
            .roi_response_series["traces"]
            .timestamps
        )
        f_interface = Fluorescence(name="corrected_fluorescence")
        ophys_module.add_data_interface(f_interface)

        f_interface.create_roi_response_series(
            name="traces",
            data=traces.T,  # Should be stored as timepoints x rois
            unit="NA",
            rois=roi_table_region,
            timestamps=ophys_timestamps,
        )

        # For the r and RMSE values we use the response series to store
        # the values as a convenience. The timestamps are thus superfluous
        # and we fill them with a dummy array of the same length of r/RMSE.
        f_interface.create_roi_response_series(
            name="r",
            data=r_values,
            unit="NA",
            rois=roi_table_region,
            timestamps=np.arange(len(r_values)),
        )

        f_interface.create_roi_response_series(
            name="RMSE",
            data=rmse,
            unit="NA",
            rois=roi_table_region,
            timestamps=np.arange(len(rmse)),
        )

        return nwbfile
