import numpy as np
import pandas as pd
from pynwb import NWBFile
from pynwb.ophys import Fluorescence

from allensdk.brain_observatory.behavior.data_files.neuropil_file import (
    NeuropilFile,
)
from allensdk.core import DataObject
from allensdk.core import \
    DataFileReadableInterface, NwbReadableInterface
from allensdk.core import \
    NwbWritableInterface
from allensdk.brain_observatory.behavior.data_objects.cell_specimens\
    .rois_mixin import \
    RoisMixin


class NeuropilTraces(
    DataObject,
    RoisMixin,
    DataFileReadableInterface,
    NwbReadableInterface,
    NwbWritableInterface,
):
    """A data container to load, access, and store the
    neuropil_traces dataframe. Neuropil traces are the fluorescent signal
    measured from the neuropil_masks.
    """

    def __init__(self, traces: pd.DataFrame):
        """
        Parameters
        ----------
        traces
            index cell_roi_id
            columns:
            - neuropil_traces
                list of float
        """
        super().__init__(name="neuropil_traces", value=traces)

    @classmethod
    def from_nwb(cls, nwbfile: NWBFile) -> "NeuropilTraces":
        # TODO Remove try/except once VBO released.
        try:
            neuropil_traces_nwb = (
                nwbfile.processing["ophys"]
                .data_interfaces["neuropil_trace"]
                .roi_response_series["traces"]
            )
            # f traces stored as timepoints x rois in NWB
            # We want rois x timepoints, hence the transpose
            f_traces = neuropil_traces_nwb.data[:].T.copy()
            roi_ids = neuropil_traces_nwb.rois.table.id[:].copy()
            df = pd.DataFrame(
                {"neuropil_trace": [x for x in f_traces]},
                index=pd.Index(data=roi_ids, name="cell_roi_id"),
            )
            return NeuropilTraces(traces=df)
        except KeyError:
            return None

    @classmethod
    def from_data_file(cls, neuropil_file: NeuropilFile) -> "NeuropilTraces":
        neuropil_traces = neuropil_file.data
        return cls(traces=neuropil_traces)

    def to_nwb(self, nwbfile: NWBFile) -> NWBFile:
        neuropil_traces = self.value["neuropil_trace"]

        # Convert from Series of lists to numpy array
        # of shape ROIs x timepoints
        traces = np.stack([x for x in neuropil_traces])

        # Create/Add neuropil_traces modules and interfaces:
        ophys_module = nwbfile.processing["ophys"]

        roi_table_region = (
            nwbfile.processing["ophys"]
            .data_interfaces["dff"]
            .roi_response_series["traces"]
            .rois
        )  # noqa: E501
        ophys_timestamps = (
            ophys_module.get_data_interface("dff")
            .roi_response_series["traces"]
            .timestamps
        )
        f_interface = Fluorescence(name="neuropil_trace")
        ophys_module.add_data_interface(f_interface)

        f_interface.create_roi_response_series(
            name="traces",
            data=traces.T,  # Should be stored as timepoints x rois
            unit="NA",
            rois=roi_table_region,
            timestamps=ophys_timestamps,
        )
        return nwbfile
