from typing import Optional

from pynwb import NWBFile

from allensdk.core import DataObject
from allensdk.brain_observatory.behavior.data_objects import BehaviorSessionId
from allensdk.core import \
    JsonReadableInterface, NwbReadableInterface, \
    LimsReadableInterface
from allensdk.brain_observatory.behavior.data_objects.metadata\
    .subject_metadata.reporter_line import \
    ReporterLine
from allensdk.brain_observatory.behavior.data_objects.timestamps \
    .ophys_timestamps import OphysTimestamps
from allensdk.brain_observatory.behavior.data_objects.timestamps.util import \
    calc_frame_rate
from allensdk.internal.api import PostgresQueryMixin


class ImagingPlane(DataObject, LimsReadableInterface,
                   JsonReadableInterface, NwbReadableInterface):
    def __init__(self, ophys_frame_rate: float,
                 targeted_structure: str,
                 excitation_lambda: float,
                 indicator: Optional[str]):
        super().__init__(name='imaging_plane', value=None,
                         is_value_self=True)
        self._ophys_frame_rate = ophys_frame_rate
        self._targeted_structure = targeted_structure
        self._excitation_lambda = excitation_lambda
        self._indicator = indicator

    @classmethod
    def from_lims(cls, ophys_experiment_id: int,
                  lims_db: PostgresQueryMixin,
                  ophys_timestamps: OphysTimestamps,
                  excitation_lambda=910.0) -> "ImagingPlane":
        behavior_session_id = BehaviorSessionId.from_lims(
            db=lims_db, ophys_experiment_id=ophys_experiment_id)
        ophys_frame_rate = calc_frame_rate(timestamps=ophys_timestamps.value)
        targeted_structure = cls._get_targeted_structure_from_lims(
            ophys_experiment_id=ophys_experiment_id, lims_db=lims_db)
        reporter_line = ReporterLine.from_lims(
            behavior_session_id=behavior_session_id.value, lims_db=lims_db)
        indicator = reporter_line.parse_indicator(warn=True)
        return cls(ophys_frame_rate=ophys_frame_rate,
                   targeted_structure=targeted_structure,
                   excitation_lambda=excitation_lambda,
                   indicator=indicator)

    @classmethod
    def from_json(cls, dict_repr: dict,
                  ophys_timestamps: OphysTimestamps,
                  excitation_lambda=910.0) -> "ImagingPlane":
        targeted_structure = dict_repr['targeted_structure']
        ophys_fame_rate = calc_frame_rate(timestamps=ophys_timestamps.value)
        reporter_line = ReporterLine.from_json(dict_repr=dict_repr)
        indicator = reporter_line.parse_indicator(warn=True)
        return cls(targeted_structure=targeted_structure,
                   ophys_frame_rate=ophys_fame_rate,
                   excitation_lambda=excitation_lambda,
                   indicator=indicator)

    @classmethod
    def from_nwb(cls, nwbfile: NWBFile) -> "ImagingPlane":
        ophys_module = nwbfile.processing['ophys']
        image_seg = ophys_module.data_interfaces['image_segmentation']
        imaging_plane = image_seg.plane_segmentations[
            'cell_specimen_table'].imaging_plane
        ophys_frame_rate = imaging_plane.imaging_rate
        targeted_structure = imaging_plane.location
        excitation_lambda = imaging_plane.excitation_lambda

        reporter_line = ReporterLine.from_nwb(nwbfile=nwbfile)
        indicator = reporter_line.parse_indicator(warn=True)
        return cls(ophys_frame_rate=ophys_frame_rate,
                   targeted_structure=targeted_structure,
                   excitation_lambda=excitation_lambda,
                   indicator=indicator)

    @property
    def ophys_frame_rate(self) -> float:
        return self._ophys_frame_rate

    @property
    def targeted_structure(self) -> str:
        return self._targeted_structure

    @property
    def excitation_lambda(self) -> float:
        return self._excitation_lambda

    @property
    def indicator(self) -> Optional[str]:
        return self._indicator

    @staticmethod
    def _get_targeted_structure_from_lims(ophys_experiment_id: int,
                                          lims_db: PostgresQueryMixin) -> str:
        query = """
                SELECT st.acronym
                FROM ophys_experiments oe
                LEFT JOIN structures st ON st.id = oe.targeted_structure_id
                WHERE oe.id = {};
                """.format(ophys_experiment_id)
        targeted_structure = lims_db.fetchone(query, strict=True)
        return targeted_structure
