import re
import warnings
from datetime import datetime
from typing import Optional

import pytz
from allensdk.brain_observatory.behavior.data_objects.metadata.behavior_metadata.date_of_acquisition import (  # noqa: E501
    DateOfAcquisition,
)
from allensdk.core import (
    DataObject,
    JsonReadableInterface,
    LimsReadableInterface,
    NwbReadableInterface,
)
from allensdk.internal.api import PostgresQueryMixin
from pynwb import NWBFile


class Age(
    DataObject,
    JsonReadableInterface,
    LimsReadableInterface,
    NwbReadableInterface,
):
    """Age (in days) of animal at the time the behavior session was taken."""

    def __init__(self, age: int):
        super().__init__(name="age_in_days", value=age)

    @classmethod
    def from_json(cls, dict_repr: dict) -> "Age":
        age = dict_repr["age"]
        age = cls._age_code_to_days(age=age)
        return cls(age=age)

    @classmethod
    def from_lims(
        cls, behavior_session_id: int, lims_db: PostgresQueryMixin
    ) -> "Age":
        date_of_acquisition = DateOfAcquisition.from_lims(
            behavior_session_id=behavior_session_id,
            lims_db=lims_db
        ).value

        query = f"""
            SELECT d.date_of_birth AS date_of_birth
            FROM behavior_sessions bs
            JOIN donors d ON d.id = bs.donor_id
            WHERE bs.id = {behavior_session_id};
        """
        date_of_birth = cls._check_timezone(
            lims_db.fetchone(query, strict=True)
        )

        age = (date_of_acquisition - date_of_birth).days
        return cls(age=age)

    @classmethod
    def _check_timezone(cls, input_date: datetime) -> datetime:
        if input_date.tzinfo is None:
            # Add UTC tzinfo if not already set
            input_date = pytz.utc.localize(input_date)
        return input_date

    @classmethod
    def from_nwb(cls, nwbfile: NWBFile) -> "Age":
        age = cls._age_code_to_days(age=nwbfile.subject.age)
        return cls(age=age)

    @staticmethod
    def to_iso8601(age: int):
        if age is None:
            return "null"
        return f"P{age}D"

    @staticmethod
    def _age_code_to_days(age: str, warn=False) -> Optional[int]:
        """Converts the age code into a numeric days representation

        Parameters
        ----------
        age
            age code, ie P123
        warn
            Whether to output warning if parsing fails
        """
        if not age.startswith("P"):
            if warn:
                warnings.warn(
                    "Could not parse numeric age from age code "
                    '(age code does not start with "P")'
                )
            return None

        match = re.search(r"\d+", age)

        if match is None:
            if warn:
                warnings.warn(
                    "Could not parse numeric age from age code "
                    "(no numeric values found in age code)"
                )
            return None

        start, end = match.span()
        return int(age[start:end])
