import numpy as np
import pandas as pd

from allensdk.brain_observatory.behavior.data_files import BehaviorStimulusFile
from allensdk.brain_observatory.behavior.data_objects import StimulusTimestamps


class FingerprintStimulus:
    """The fingerprint stimulus is a movie used to trigger many neurons
        and is used to improve cell matching."""
    def __init__(self, table: pd.DataFrame):
        self._table = table

    @property
    def table(self) -> pd.DataFrame:
        """Returns the fingerprint stimulus table"""
        return self._table

    @classmethod
    def from_stimulus_file(
            cls,
            stimulus_presentations: pd.DataFrame,
            stimulus_file: BehaviorStimulusFile,
            stimulus_timestamps: StimulusTimestamps
    ) -> "FingerprintStimulus":
        """
        Instantiates `FingerprintStimulus` from stimulus file

        Parameters
        ----------
        stimulus_presentations:
            Table containing previous stimuli
        stimulus_file
            BehaviorStimulusFile
        stimulus_timestamps
            StimulusTimestamps

        Returns
        -------
        `FingerprintStimulus`
            Instantiated FingerprintStimulus
        """
        fingerprint_stim = (
            stimulus_file.data['items']['behavior']['items']['fingerprint']
            ['static_stimulus'])

        n_repeats = fingerprint_stim['runs']

        # spontaneous + fingerprint indices relative to start of session
        stimulus_session_frame_indices = np.array(
            stimulus_file.data['items']['behavior']['items']
            ['fingerprint']['frame_indices'])

        movie_length = int(len(fingerprint_stim['sweep_frames']) / n_repeats)

        # Start index within the spontaneous + fingerprint block
        movie_start_index = (fingerprint_stim['frame_list'] == -1).sum()

        res = []
        for repeat in range(n_repeats):
            for frame in range(movie_length):
                # 0-indexed frame indices relative to start of fingerprint
                # movie
                stimulus_frame_indices = \
                    np.array(fingerprint_stim['sweep_frames']
                             [frame + (repeat * movie_length)])
                start_frame, end_frame = stimulus_session_frame_indices[
                    stimulus_frame_indices + movie_start_index]
                start_time, stop_time = \
                    stimulus_timestamps.value[[
                        start_frame,
                        # Sometimes stimulus timestamps gets truncated too
                        # early. There should be 2 extra frames after last
                        # stimulus presentation frame, since if the end
                        # frame is end_frame, then the end timestamp occurs on
                        # end_frame+1. The min is being taken to prevent
                        # index out of bounds. This results in the last
                        # frame's duration being too short TODO this is
                        #  probably a bug somewhere in timestamp creation
                        min(end_frame + 1,
                            len(stimulus_timestamps.value) - 1)]]
                res.append({
                    'movie_frame_index': frame,
                    'start_time': start_time,
                    'stop_time': stop_time,
                    'start_frame': start_frame,
                    'end_frame': end_frame,
                    'movie_repeat': repeat,
                    'duration': stop_time - start_time
                })
        table = pd.DataFrame(res)

        table['stimulus_block'] = \
            stimulus_presentations['stimulus_block'].max() \
            + 2     # + 2 since there is a gap before this stimulus
        table['stimulus_name'] = 'natural_movie_one'

        table = table.astype(
            {c: 'int64' for c in table.select_dtypes(include='int')})

        return FingerprintStimulus(table=table)
