from typing import Optional, Union, List

import numpy as np
from pynwb import NWBFile, ProcessingModule
from pynwb.base import TimeSeries

from allensdk.core import \
    LimsReadableInterface, NwbReadableInterface, JsonReadableInterface
from allensdk.brain_observatory.behavior.data_files.sync_file import \
    SyncFileReadableInterface
from allensdk.brain_observatory.behavior.data_files.stimulus_file import \
    StimulusFileReadableInterface
from allensdk.core import DataObject
from allensdk.brain_observatory.behavior.data_files import (
    BehaviorStimulusFile,
    MappingStimulusFile,
    ReplayStimulusFile,
    SyncFile
)
from allensdk.core import NwbWritableInterface
from allensdk.brain_observatory.behavior.data_objects.timestamps\
    .stimulus_timestamps.timestamps_processing import (
        get_behavior_stimulus_timestamps, get_ophys_stimulus_timestamps)
from allensdk.internal.api import PostgresQueryMixin
from allensdk.brain_observatory.sync_stim_aligner import (
    get_stim_timestamps_from_stimulus_blocks)


class StimulusTimestamps(DataObject,
                         StimulusFileReadableInterface,
                         SyncFileReadableInterface,
                         NwbReadableInterface,
                         LimsReadableInterface,
                         NwbWritableInterface,
                         JsonReadableInterface):
    """A DataObject which contains properties and methods to load, process,
    and represent visual behavior stimulus timestamp data.

    Stimulus timestamp data is represented as:

    Numpy array whose length is equal to the number of timestamps collected
    and whose values are timestamps (in seconds)
    """

    def __init__(
        self,
        timestamps: np.ndarray,
        monitor_delay: float,
        stimulus_file: Optional[BehaviorStimulusFile] = None,
        sync_file: Optional[SyncFile] = None
    ):
        super().__init__(name="stimulus_timestamps",
                         value=timestamps+monitor_delay)
        self._stimulus_file = stimulus_file
        self._sync_file = sync_file
        self._monitor_delay = monitor_delay

    def update_timestamps(
            self,
            timestamps: np.ndarray
    ) -> "StimulusTimestamps":
        """
        Returns newly instantiated `StimulusTimestamps` with `timestamps`

        Parameters
        ----------
        timestamps

        Returns
        -------
        `StimulusTimestamps` with `timestamps`
        """
        return StimulusTimestamps(
            timestamps=timestamps,
            monitor_delay=self._monitor_delay,
            stimulus_file=self._stimulus_file,
            sync_file=self._sync_file
        )

    def subtract_monitor_delay(self) -> "StimulusTimestamps":
        """
        Return a version of this StimulusTimestamps object with
        monitor_delay = 0 by subtracting self.monitor_delay from
        self.value
        """
        new_value = self.value-self.monitor_delay
        return StimulusTimestamps(
                    timestamps=new_value,
                    monitor_delay=0.0)

    @property
    def monitor_delay(self) -> float:
        return self._monitor_delay

    @classmethod
    def from_stimulus_file(
            cls,
            stimulus_file: BehaviorStimulusFile,
            monitor_delay: float) -> "StimulusTimestamps":
        stimulus_timestamps = get_behavior_stimulus_timestamps(
            stimulus_pkl=stimulus_file.data
        )

        return cls(
            timestamps=stimulus_timestamps,
            monitor_delay=monitor_delay,
            stimulus_file=stimulus_file
        )

    @classmethod
    def from_sync_file(
            cls,
            sync_file: SyncFile,
            monitor_delay: float) -> "StimulusTimestamps":
        stimulus_timestamps = get_ophys_stimulus_timestamps(
            sync_path=sync_file.filepath
        )
        return cls(
            timestamps=stimulus_timestamps,
            monitor_delay=monitor_delay,
            sync_file=sync_file
        )

    @classmethod
    def from_json(
            cls,
            dict_repr: dict,
            monitor_delay=0.0
    ) -> "StimulusTimestamps":
        """
        Reads timestamps from stimulus file or sync file.
        Note that `from_multiple_stimulus_blocks` method of constructing
        timestamps is not supported using `from_json`

        Parameters
        ----------
        dict_repr
        monitor_delay: Monitor delay to apply to the timestamps

        Returns
        -------
        StimulusTimestamps
        """
        if 'sync_file' in dict_repr:
            sync_file = SyncFile.from_json(dict_repr=dict_repr)
            return cls.from_sync_file(
                        sync_file=sync_file,
                        monitor_delay=monitor_delay)
        else:
            stim_file = BehaviorStimulusFile.from_json(dict_repr=dict_repr)
            return cls.from_stimulus_file(
                        stimulus_file=stim_file,
                        monitor_delay=monitor_delay)

    @classmethod
    def from_multiple_stimulus_blocks(
            cls,
            sync_file: SyncFile,
            list_of_stims: List[Union[BehaviorStimulusFile,
                                      MappingStimulusFile,
                                      ReplayStimulusFile]],
            stims_of_interest: Optional[List[int]] = None,
            monitor_delay: float = 0.0,
            frame_time_lines: Union[str, List[str]] = 'vsync_stim',
            frame_time_line_direction: str = 'rising',
            frame_count_tolerance: float = 0.0) -> "StimulusTimestamps":
        """
        Construct a StimulusTimestamps instance by registering
        multiple stimulus blocks to one sync file and concatenating the results

        Parameters
        ----------
        sync_file: SyncFile

        list_of_stims: List[Union[BehaviorStimulusFile,
                                  MappingStimulusFile,
                                  ReplayStimulusFile]]
            The list of StimulusFiles to be registered to the SyncFile
            **in the order that they were presented to the mouse**

        stims_of_interest: Optional[List[int]]
            The indexes in list_of_stims of the timestamps to be
            concatenated into this one StimulusTimestamps object.
            If `None` (default), the timestamps from all stimulus files
            are concatenated

        monitor_delay: float
            in seconds

        frame_time_lines: Union[str, List[str]]
            The line to be used to find raw frame times (usually 'vsync_stim').
            If a list, the code will scan the list in order until a line
            that is present in the sync file is found. That line will be used.

        frame_time_line_direction: str
            Either 'rising' or 'falling' indicating which edge
            to use in finding the raw frame times

        frame_count_tolerance: float
            The tolerance to within two blocks of frame counts are considered
            equal
        """
        behavior_stimulus_files = [x for x in list_of_stims
                                   if isinstance(x, BehaviorStimulusFile)]
        if len(behavior_stimulus_files) == 0:
            raise ValueError(
                'One of the values in `list_of_stims` must be a '
                '`BehaviorStimulusFile`')
        elif len(behavior_stimulus_files) > 1:
            raise ValueError('You passed multiple `BehaviorStimulusFile` to '
                             '`list_of_stims`. Please pass only 1.')

        if stims_of_interest:
            if len(stims_of_interest) > len(list_of_stims):
                raise ValueError(
                    f'stims_of_interest has length {len(stims_of_interest)} '
                    f'but list_of_stims has length {len(list_of_stims)}. '
                    f'len(stims_of_interest) should be <= len(list_of_stims)')
            if any([x < 0 for x in stims_of_interest]):
                raise ValueError('stims_of_interest should not be negative')
            if any([x >= len(list_of_stims) for x in stims_of_interest]):
                raise ValueError('stims_of_interest contains an index '
                                 'greater than the number of elements in '
                                 'list_of_stims')
        stimulus_times = get_stim_timestamps_from_stimulus_blocks(
                        stimulus_files=list_of_stims,
                        sync_file=sync_file.filepath,
                        raw_frame_time_lines=frame_time_lines,
                        raw_frame_time_direction=frame_time_line_direction,
                        frame_count_tolerance=frame_count_tolerance)

        to_concatenate = \
            [t for t in stimulus_times["timestamps"]] \
            if stims_of_interest is None else  \
            [stimulus_times["timestamps"][idx] for idx in stims_of_interest]

        timestamps = np.concatenate(to_concatenate)

        return cls(timestamps=timestamps,
                   monitor_delay=monitor_delay,
                   sync_file=sync_file,
                   stimulus_file=behavior_stimulus_files[0])

    def from_lims(
        cls,
        db: PostgresQueryMixin,
        monitor_delay: float,
        behavior_session_id: int,
        ophys_experiment_id: Optional[int] = None
    ) -> "StimulusTimestamps":
        stimulus_file = BehaviorStimulusFile.from_lims(
                            db,
                            behavior_session_id)

        if ophys_experiment_id:
            sync_file = SyncFile.from_lims(
                db=db, ophys_experiment_id=ophys_experiment_id)
            return cls.from_sync_file(sync_file=sync_file,
                                      monitor_delay=monitor_delay)
        else:
            return cls.from_stimulus_file(stimulus_file=stimulus_file,
                                          monitor_delay=monitor_delay)

    @classmethod
    def from_nwb(cls,
                 nwbfile: NWBFile) -> "StimulusTimestamps":
        stim_module = nwbfile.processing["stimulus"]
        stim_ts_interface = stim_module.get_data_interface("timestamps")
        stim_timestamps = stim_ts_interface.timestamps[:]

        # Because the monitor delay was already applied when
        # saving the stimulus timestamps to the NWB file,
        # we set it to zero here.
        return cls(timestamps=stim_timestamps,
                   monitor_delay=0.0)

    def to_nwb(self, nwbfile: NWBFile) -> NWBFile:
        stimulus_ts = TimeSeries(
            data=self._value,
            name="timestamps",
            timestamps=self._value,
            unit="s"
        )

        stim_mod = ProcessingModule("stimulus", "Stimulus Times processing")
        stim_mod.add_data_interface(stimulus_ts)
        nwbfile.add_processing_module(stim_mod)

        return nwbfile

    def __len__(self):
        return len(self.value)
