from typing import Union
from pathlib import Path

import numpy as np

from allensdk.internal.brain_observatory.time_sync import OphysTimeAligner


def get_behavior_stimulus_timestamps(stimulus_pkl: dict) -> np.ndarray:
    """Obtain visual behavior stimuli timing information from a behavior
    stimulus *.pkl file.

    Parameters
    ----------
    stimulus_pkl : dict
        A dictionary containing stimulus presentation timing information
        during a behavior session. Presentation timing info is stored as
        an array of times between frames (frame time intervals) in
        milliseconds.

    Returns
    -------
    np.ndarray
        Timestamps (in seconds) for presented stimulus frames during a session.
    """
    vsyncs = stimulus_pkl["items"]["behavior"]["intervalsms"]
    stimulus_timestamps = np.hstack((0, vsyncs)).cumsum() / 1000.0
    return stimulus_timestamps


def get_ophys_stimulus_timestamps(sync_path: Union[str, Path]) -> np.ndarray:
    """Obtain visual behavior stimuli timing information from a sync *.h5 file.

    Parameters
    ----------
    sync_path : Union[str, Path]
        The path to a sync *.h5 file that contains global timing information
        about multiple data streams (e.g. behavior, ophys, eye_tracking)
        during a session.

    Returns
    -------
    np.ndarray
        Timestamps (in seconds) for presented stimulus frames during a
        behavior + ophys session.
    """
    aligner = OphysTimeAligner(sync_file=sync_path)
    stimulus_timestamps, _ = aligner.clipped_stim_timestamps
    return stimulus_timestamps


def get_frame_indices(
        frame_timestamps: np.ndarray,
        event_timestamps: np.ndarray) -> np.ndarray:
    """
    Given an array of timestamps corresponding to stimulus frames
    and an array of timestamps corresponding to some event (i.e.
    licks), return an array of indexes indicating which frame
    each event occured on. Indexes will be chosen to be the
    first index satisfying

    frame_timestamps[event_indices] >= event_timestamps
    event_timestamps < frame_timestamps[event_indices+1]

    Parameters
    ----------
    frame_timestamps: np.ndarray
        must be in ascending order

    event_timetamps: np.ndarray

    Returns
    -------
    event_indices: np.ndarray
        integers that are all >= 0 and < len(frame_timestamps)

    Note
    ----
    If a value is repeated in frame_timestamps, the first suitable
    frame index is returned for a given event_timestamp
    """

    if np.any(np.diff(frame_timestamps) < -1.0e-10):
        raise ValueError("frame_timestamps are not in ascending order")

    n_frames = len(frame_timestamps)

    event_indices = np.searchsorted(
                        frame_timestamps,
                        event_timestamps,
                        side='left')

    event_indices = np.clip(event_indices, None, n_frames-1)

    # correct for fact that searchsorted will select as
    # frame index the first frame time that is larger
    # than lick_times; we want the lick_times associated
    # with the last frame that is smaller than the lick_time
    event_frame_times = frame_timestamps[event_indices]
    delta = event_timestamps-event_frame_times
    to_decrement = (delta < -1.0e-6)
    event_indices[to_decrement] -= 1
    event_indices = np.clip(event_indices, 0, n_frames-1)

    return event_indices
