import logging
from pathlib import Path
from typing import List

import argschema
from allensdk.brain_observatory.behavior.behavior_ophys_experiment import (
    BehaviorOphysExperiment,
)
from allensdk.brain_observatory.behavior.write_nwb.nwb_writer_utils import (
    OphysNwbWriter,
)
from allensdk.brain_observatory.behavior.write_nwb.ophys.schemas import (
    OphysExperimentInputSchema,
    OutputSchema,
)


class WriteOphysNWB(argschema.ArgSchemaParser):
    default_schema = OphysExperimentInputSchema
    default_output_schema = OutputSchema

    def run(self):
        self.logger.name = type(self).__name__
        oe_id = self.args["ophys_experiment_id"]
        oe_id_dir = Path(self.args["output_dir_path"]) / f"{oe_id}"
        oe_id_dir.mkdir(exist_ok=True)
        output_file = self.write_experiment_nwb(
            ophys_experiment_metadata=self.args["ophys_experiment_metadata"],
            nwb_filepath=oe_id_dir / f"behavior_ophys_experiment_{oe_id}.nwb",
            skip_metadata=self.args["skip_metadata_key"],
            skip_stim=self.args["skip_stimulus_file_key"],
        )
        logging.info("File successfully created")

        output_dict = {"output_path": output_file,
                       "input_parameters": self.args}

        self.output(output_dict)

    def write_experiment_nwb(
        self,
        ophys_experiment_metadata: dict,
        nwb_filepath: Path,
        skip_metadata: List[str],
        skip_stim: List[str],
    ) -> str:
        """Load and write a BehaviorOphysExperiment as NWB.

        Check data object against associated data from the ophys_experiment
        metadata table. Additionally, round trip the NWB file to confirm that
        it was properly created for the session.

        Parameters
        ----------
        ophys_experiment_metadata : dict
            Dictionary of keys that overlap between the metadata in the
            BehaviorOphysExperiment object and the associated metadata table.
        nwb_filepath : pathlib.Path
            Base filename path to write the NWB to.
        skip_metadata : list of str
            List of metadata keys to skip when comparing data.
        skip_stim : list of str
            List of stimulus file keys to skip when comparing data.

        Returns
        -------
        output_path : str
            String path of where the NWB was written to.

        Note
        ----
        Upon encountering an error, either from write or data comparison, the
        code changes the file name suffix to error and does not return the
        path of the file.
        """
        nwb_writer = OphysNwbWriter(
            nwb_filepath=str(nwb_filepath),
            session_data=ophys_experiment_metadata,
            serializer=BehaviorOphysExperiment,
        )
        nwb_writer.write_nwb(
            id_column_name="ophys_experiment_id",
            ophys_experiment_ids=self.args["ophys_container_experiment_ids"],
            skip_metadata=skip_metadata,
            skip_stim=skip_stim,
        )

        return str(nwb_filepath)


if __name__ == "__main__":
    nwb_writer = WriteOphysNWB()
    nwb_writer.run()
