from typing import Optional
import numpy as np

from allensdk.core import DataObject
from allensdk.brain_observatory.ecephys.utils import (
    strip_substructure_acronym)


class Channel(DataObject):
    """Probe channel"""
    def __init__(
            self,
            id: int,
            probe_id: int,
            valid_data: bool,
            probe_channel_number: int,
            probe_vertical_position: int,
            probe_horizontal_position: int,
            structure_acronym: str = '',
            anterior_posterior_ccf_coordinate: Optional[float] = None,
            dorsal_ventral_ccf_coordinate: Optional[float] = None,
            left_right_ccf_coordinate: Optional[float] = None,
            impedance: float = np.nan,
            filtering: str = 'AP band: 500 Hz high-pass; '
                             'LFP band: 1000 Hz low-pass',
            strip_structure_subregion: bool = True
    ):
        """

        Parameters
        ----------
        strip_structure_subregion: Whether to remove the subregion from the
            structure acronym. I.e if the acronym is "LGd-sh" then it will get
            parsed as "LGd". You might want to strip it if the subregion is
            beyond annotation accuracy.
        """
        super().__init__(name='channel',
                         value=None,
                         is_value_self=True)
        self._id = id
        self._probe_id = probe_id
        self._valid_data = valid_data
        self._probe_channel_number = probe_channel_number
        self._probe_vertical_position = probe_vertical_position
        self._probe_horizontal_position = probe_horizontal_position
        self._structure_acronym = structure_acronym
        self._anterior_posterior_ccf_coordinate = \
            anterior_posterior_ccf_coordinate
        self._dorsal_ventral_ccf_coordinate = dorsal_ventral_ccf_coordinate
        self._left_right_ccf_coordinate = left_right_ccf_coordinate
        self._impedance = impedance
        self._filtering = filtering
        self._strip_structure_subregion = strip_structure_subregion

    @property
    def id(self) -> int:
        return self._id

    @property
    def probe_id(self) -> int:
        return self._probe_id

    @property
    def valid_data(self) -> bool:
        return self._valid_data

    @property
    def probe_channel_number(self) -> int:
        return self._probe_channel_number

    @property
    def probe_vertical_position(self) -> int:
        return self._probe_vertical_position

    @property
    def probe_horizontal_position(self) -> int:
        return self._probe_horizontal_position

    @property
    def structure_acronym(self) -> str:
        acronym = self._structure_acronym
        if self._strip_structure_subregion:
            acronym = strip_substructure_acronym(self._structure_acronym)
        return acronym

    @property
    def anterior_posterior_ccf_coordinate(self) -> Optional[float]:
        return self._anterior_posterior_ccf_coordinate

    @property
    def dorsal_ventral_ccf_coordinate(self) -> Optional[float]:
        return self._dorsal_ventral_ccf_coordinate

    @property
    def left_right_ccf_coordinate(self) -> Optional[float]:
        return self._left_right_ccf_coordinate

    @property
    def impedance(self) -> float:
        return self._impedance

    @property
    def filtering(self) -> str:
        return self._filtering
