from typing import List, Optional

import pandas as pd
from pynwb import NWBFile

from allensdk.brain_observatory.ecephys._channel import Channel
from allensdk.brain_observatory.ecephys.utils import clobbering_merge
from allensdk.core import DataObject, NwbReadableInterface, \
    JsonReadableInterface


class Channels(DataObject, NwbReadableInterface, JsonReadableInterface):
    """A set of channels"""

    def __init__(self, channels: List[Channel]):
        super().__init__(name='channels', value=channels)

    @classmethod
    def from_json(
            cls,
            channels: dict
    ) -> "Channels":
        for channel in channels:
            if 'impedence' in channel:
                # Correct misspelling
                channel['impedance'] = channel.pop('impedence')

        channels = [Channel(
            id=channel['id'],
            probe_id=channel['probe_id'],
            valid_data=channel['valid_data'],
            probe_channel_number=channel['probe_channel_number'],
            probe_vertical_position=channel['probe_vertical_position'],
            probe_horizontal_position=channel['probe_horizontal_position'],
            structure_acronym=channel['structure_acronym'],
            anterior_posterior_ccf_coordinate=(
                channel['anterior_posterior_ccf_coordinate']),
            dorsal_ventral_ccf_coordinate=(
                channel['dorsal_ventral_ccf_coordinate']),
            left_right_ccf_coordinate=channel['left_right_ccf_coordinate']
        )
                    for channel in channels]
        return Channels(channels=channels)

    def to_dataframe(self, external_channel_columns=None,
                     filter_by_validity=True) -> pd.DataFrame:
        """

        Parameters
        ----------
        external_channel_columns
        filter_by_validity: Whether to filter channels based on whether
            the channel is marked as "valid_data"

        Returns
        -------

        """
        channels = [channel.to_dict()['channel'] for channel in self.value]
        channels = pd.DataFrame(channels)
        channels = channels.set_index('id')
        channels = channels.drop(columns=['impedance'])

        if external_channel_columns is not None:
            external_channel_columns = external_channel_columns()
            channels = clobbering_merge(channels, external_channel_columns,
                                        left_index=True, right_index=True)

        if filter_by_validity:
            channels = channels[channels['valid_data']]
            channels = channels.drop(columns=['valid_data'])
        return channels

    @classmethod
    def from_nwb(cls, nwbfile: NWBFile,
                 probe_id: Optional[int] = None) -> "Channels":
        """

        Parameters
        ----------
        nwbfile
        probe_id: Filter channels to only those for `probe_id`

        Returns
        -------
        Channels
        """
        channels = []
        for channel_id, row in nwbfile.electrodes.to_dataframe().iterrows():
            if probe_id is not None:
                if row['probe_id'] != probe_id:
                    continue

            # this block of code is necessary to maintain
            # backwards compatibility with Visual Coding Neuropixels
            # NWB files, which used 'local_index' to mean what we
            # now mean by 'probe_channel_number'
            has_local_index = ('local_index' in row.keys())
            has_channel_number = ('probe_channel_number' in row.keys())
            if has_local_index and has_channel_number:
                raise RuntimeError("Unclear how to read channel; "
                                   "has both 'local_index' and "
                                   "'probe_channel_number'")
            elif has_local_index:
                idx_col = 'local_index'
            elif has_channel_number:
                idx_col = 'probe_channel_number'
            else:
                raise RuntimeError("Unclear how to read channel; "
                                   "has neither 'local_index' nor "
                                   "'probe_channel_number'.\n"
                                   f"Columns are {row.keys()}")

            structure_acronym = \
                None if row['location'] in ['None', ''] else row['location']
            channels.append(Channel(
                id=channel_id,
                probe_channel_number=row[idx_col],
                probe_horizontal_position=row['probe_horizontal_position'],
                probe_vertical_position=row['probe_vertical_position'],
                probe_id=row['probe_id'],
                valid_data=row['valid_data'],
                structure_acronym=structure_acronym,
                anterior_posterior_ccf_coordinate=row['x'],
                dorsal_ventral_ccf_coordinate=row['y'],
                left_right_ccf_coordinate=row['z']
            ))
        return Channels(channels=channels)
