import dataclasses
import logging
from datetime import datetime
from pathlib import Path
from typing import Optional, Union, Callable, Tuple

import numpy as np
import pandas as pd
import pynwb
from pynwb import NWBFile
from xarray import DataArray

from allensdk.brain_observatory.ecephys._behavior_ecephys_metadata import \
    BehaviorEcephysMetadata
from allensdk.brain_observatory.ecephys._channels import Channels
from allensdk.brain_observatory.ecephys._current_source_density import \
    CurrentSourceDensity
from allensdk.brain_observatory.ecephys._units import Units
from allensdk.brain_observatory.ecephys._lfp import LFP
from allensdk.brain_observatory.ecephys.nwb import EcephysCSD
from allensdk.brain_observatory.ecephys.nwb_util import add_probe_to_nwbfile, \
    add_ecephys_electrodes
from allensdk.core import DataObject, JsonReadableInterface, \
    NwbWritableInterface, NwbReadableInterface


@dataclasses.dataclass
class ProbeWithLFPMeta:
    """
    Metadata for a single probe which has LFP data associated with it

    Attributes:

    - lfp_csd_filepath --> Either a path to the NWB file containing the LFP 
    and CSD data or a callable which returns it. The nwb file is loaded 
    separately from the main session nwb file in order to load the LFP data 
    on the fly rather than with the main session NWB file. This is to speed 
    up download of the NWB for users who don't wish to load the LFP data (it 
    is large).
    - lfp_sampling_rate --> LFP sampling rate
    """  # noqa E402

    lfp_csd_filepath: Union[Path, Callable[[], Path]]
    lfp_sampling_rate: float


class Probe(DataObject, JsonReadableInterface, NwbWritableInterface,
            NwbReadableInterface):
    """A single probe"""
    def __init__(
            self,
            id: int,
            name: str,
            channels: Channels,
            units: Units,
            sampling_rate: float = 30000.0,
            lfp: Optional[LFP] = None,
            lfp_meta: Optional[ProbeWithLFPMeta] = None,
            current_source_density: Optional[CurrentSourceDensity] = None,
            location: str = 'See electrode locations',
            temporal_subsampling_factor: Optional[float] = 2.0
    ):
        """

        Parameters
        ----------
        id:
            probe id
        name:
            probe name
        channels:
            probe channels
        units:
            units detected by probe
        sampling_rate:
            probe sampling rate
        lfp:
            probe LFP
        lfp_meta
            `ProbeWithLFPMeta`
        current_source_density
            probe current source density
        location:
            probe location
        temporal_subsampling_factor:
            subsampling factor applied to lfp data (across time)
        """
        self._id = id
        self._name = name
        self._channels = channels
        self._units = units
        self._sampling_rate = sampling_rate
        self._lfp = lfp
        self._lfp_meta = lfp_meta
        self._current_source_density = current_source_density
        self._location = location
        self._temporal_subsampling_factor = temporal_subsampling_factor
        super().__init__(name=name,
                         value=None,
                         is_value_self=True)

    @property
    def id(self) -> int:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def channels(self) -> Channels:
        return self._channels

    @property
    def units(self) -> Units:
        return self._units

    @property
    def sampling_rate(self) -> float:
        return self._sampling_rate

    @property
    def lfp(self) -> Optional[DataArray]:
        if self._lfp is None:
            if self._lfp_meta is None:
                return None
            lfp = self._read_lfp_from_nwb()
            self._lfp = lfp

        lfp = self._lfp.to_dataarray()
        return lfp

    @property
    def current_source_density(self) -> Optional[DataArray]:
        if self._current_source_density is None:
            if self._lfp_meta is None:
                return None
            csd = self._read_csd_data_from_nwb()
            self._current_source_density = csd

        csd = self._current_source_density.to_dataarray()
        return csd

    @property
    def location(self) -> str:
        return self._location

    @property
    def temporal_subsampling_factor(self) -> Optional[float]:
        return self._temporal_subsampling_factor

    @property
    def units_table(self) -> pd.DataFrame:
        df = pd.DataFrame(
                [unit.to_dict()['unit'] for unit in self.units.value])
        df = df.fillna(np.nan)
        return df

    @classmethod
    def from_json(cls, probe: dict) -> "Probe":
        channels = Channels.from_json(channels=probe['channels'])
        units = Units.from_json(probe=probe)

        if probe['lfp'] is not None:
            lfp = LFP.from_json(
                probe_meta=probe
            )
            csd = CurrentSourceDensity.from_json(
                probe_meta=probe
            )
        else:
            lfp = None
            csd = None

        return Probe(
            id=probe['id'],
            name=probe['name'],
            channels=channels,
            units=units,
            sampling_rate=probe['sampling_rate'],
            lfp=lfp,
            current_source_density=csd,
            temporal_subsampling_factor=probe['temporal_subsampling_factor']
        )

    @classmethod
    def from_nwb(
            cls,
            nwbfile: NWBFile,
            probe_name: str,
            lfp_meta: Optional[ProbeWithLFPMeta] = None
    ) -> "Probe":
        """

        Parameters
        ----------
        nwbfile
        probe_name
            Probe name
        lfp_meta
            `ProbeWithLFPMeta`

        Returns
        -------
        `NWBFile` with probe data added
        """
        probe = nwbfile.electrode_groups[probe_name]
        channels = Channels.from_nwb(nwbfile=nwbfile, probe_id=probe.probe_id)
        units = Units.from_nwb(nwbfile=nwbfile, probe_id=probe.probe_id)
        return Probe(
            id=probe.probe_id,
            name=probe.name,
            location=probe.location,
            sampling_rate=probe.device.sampling_rate,
            channels=channels,
            units=units,
            lfp_meta=lfp_meta
        )

    def to_nwb(
            self,
            nwbfile: NWBFile
    ) -> Tuple[NWBFile, Optional[NWBFile]]:
        """

        Parameters
        ----------
        nwbfile

        Returns
        -------
        (session `NWBFile` instance,
        probe `NWBFile` instance if LFP data exists else None)
        """
        nwbfile = self._add_probe_to_nwb(nwbfile=nwbfile)

        if self._lfp is not None:
            probe_nwbfile = self.add_lfp_to_nwb(
                session_id=nwbfile.session_id,
                session_metadata=BehaviorEcephysMetadata.from_nwb(
                    nwbfile=nwbfile),
                session_start_time=nwbfile.session_start_time
            )
        else:
            logging.info(f'No LFP data found for probe {self._id}')
            probe_nwbfile = None
        return nwbfile, probe_nwbfile

    def _add_probe_to_nwb(
            self, nwbfile: NWBFile,
            add_only_lfp_channels: bool = False
    ):
        logging.info(f'found probe {self._id} with name {self._name}')

        nwbfile, probe_nwb_device, probe_nwb_electrode_group = \
            add_probe_to_nwbfile(
                nwbfile,
                probe_id=self._id,
                name=self._name,
                sampling_rate=self._sampling_rate,
                lfp_sampling_rate=(
                    self._lfp.sampling_rate if self._lfp is not None
                    else np.nan),
                has_lfp_data=self._lfp is not None
            )

        channels = [c.to_dict()['channel'] for c in self._channels.value]

        if self._lfp is not None and add_only_lfp_channels:
            channel_number_whitelist = self._lfp.channels
        else:
            channel_number_whitelist = None

        add_ecephys_electrodes(
            nwbfile,
            channels,
            probe_nwb_electrode_group,
            channel_number_whitelist=channel_number_whitelist)
        return nwbfile

    def add_lfp_to_nwb(
            self,
            session_id: str,
            session_start_time: datetime,
            session_metadata: BehaviorEcephysMetadata
    ):
        logging.info(f'writing lfp file for probe {self._id}')

        nwbfile = pynwb.NWBFile(
            session_description='LFP data and associated info for one probe',
            identifier=f"{self._id}",
            session_id=f"{session_id}",
            session_start_time=session_start_time,
            institution="Allen Institute for Brain Science"
        )
        session_metadata.to_nwb(nwbfile=nwbfile)

        nwbfile = self._add_probe_to_nwb(
            nwbfile=nwbfile,
            add_only_lfp_channels=True
        )
        lfp_nwb = pynwb.ecephys.LFP(name=f"probe_{self._id}_lfp")

        electrode_table_region = nwbfile.create_electrode_table_region(
            region=np.arange(len(nwbfile.electrodes)).tolist(),
            name='electrodes',
            description=f"lfp channels on probe {self._id}"
        )

        nwbfile.add_acquisition(lfp_nwb.create_electrical_series(
            name=f"probe_{self._id}_lfp_data",
            data=self._lfp.data,
            timestamps=self._lfp.timestamps,
            electrodes=electrode_table_region
        ))
        nwbfile.add_acquisition(lfp_nwb)

        if self._current_source_density is not None:
            nwbfile = self._add_csd_to_nwb(nwbfile=nwbfile)

        return nwbfile

    def _add_csd_to_nwb(
            self,
            nwbfile: NWBFile,
            csd_unit: str = 'V/cm^2',
            position_unit: str = "um"
    ):
        """

        Parameters
        ----------
        nwbfile:
            NWBFile containing LFP data
        csd_unit:
            Units of CSD data, by default "V/cm^2"
        position_unit:
            Units of virtual channel locations, by default "um" (micrometer)

        Returns
        -------
        `NWBFile` with csd added
        """
        csd = self._current_source_density

        csd_mod = pynwb.ProcessingModule("current_source_density",
                                         "Precalculated current source "
                                         "density")
        nwbfile.add_processing_module(csd_mod)

        csd_ts = pynwb.base.TimeSeries(
            name="current_source_density",
            data=csd.data.T,
            # TimeSeries should have data in (time x channels) format
            timestamps=csd.timestamps.T,
            unit=csd_unit
        )

        x_locs, y_locs = np.split(csd.channel_locations.astype(np.uint64),
                                  2,
                                  axis=1)

        csd = EcephysCSD(name="ecephys_csd",
                         time_series=csd_ts,
                         virtual_electrode_x_positions=x_locs.flatten(),
                         virtual_electrode_x_positions__unit=position_unit,
                         virtual_electrode_y_positions=y_locs.flatten(),
                         virtual_electrode_y_positions__unit=position_unit)

        csd_mod.add_data_interface(csd)

        return nwbfile

    def _read_lfp_from_nwb(self) -> LFP:
        if isinstance(self._lfp_meta.lfp_csd_filepath, Callable):
            logging.info('Fetching LFP NWB file')
            path = self._lfp_meta.lfp_csd_filepath()
        else:
            path = self._lfp_meta.lfp_csd_filepath
        with pynwb.NWBHDF5IO(path, 'r', load_namespaces=True) as f:
            nwbfile = f.read()
            probe = nwbfile.electrode_groups[self._name]
            lfp = nwbfile.get_acquisition(f'probe_{self._id}_lfp')
            series = lfp.get_electrical_series(f'probe_{self._id}_lfp_data')

            electrodes = nwbfile.electrodes.to_dataframe()

            data = series.data[:]
            timestamps = series.timestamps[:]

        return LFP(
            data=data,
            timestamps=timestamps,
            channels=electrodes.index.values,
            sampling_rate=probe.lfp_sampling_rate
        )

    def _read_csd_data_from_nwb(self) -> CurrentSourceDensity:
        if isinstance(self._lfp_meta.lfp_csd_filepath, Callable):
            logging.info('Fetching LFP NWB file')
            path = self._lfp_meta.lfp_csd_filepath()
        else:
            path = self._lfp_meta.lfp_csd_filepath
        with pynwb.NWBHDF5IO(path, 'r', load_namespaces=True) as f:
            nwbfile = f.read()
            csd_mod = nwbfile.get_processing_module(
                "current_source_density")
            nwb_csd = csd_mod["ecephys_csd"]
            csd_data = nwb_csd.time_series.data[:]

            # csd data stored as (timepoints x channels) but we
            # want (channels x timepoints)
            csd_data = csd_data.T

            channel_locations = np.stack([
                nwb_csd.virtual_electrode_x_positions,
                nwb_csd.virtual_electrode_y_positions], axis=1).astype('int')
            return CurrentSourceDensity(
                data=csd_data,
                timestamps=nwb_csd.time_series.timestamps[:],
                interpolated_channel_locations=channel_locations
            )

    def to_dict(self) -> dict:
        has_lfp_data = False
        lfp_sampling_rate = None

        if self._lfp is not None:
            lfp_sampling_rate = self._lfp.sampling_rate
            has_lfp_data = True
        elif self._lfp_meta is not None:
            lfp_sampling_rate = self._lfp_meta.lfp_sampling_rate
            has_lfp_data = True

        return {
            'id': self._id,
            'name': self._name,
            'location': self._location,
            'sampling_rate': self._sampling_rate,
            'lfp_sampling_rate': lfp_sampling_rate,
            'has_lfp_data': has_lfp_data
        }
