from typing import Optional, Tuple

import numpy as np

from allensdk.core import DataObject


class Unit(DataObject):
    """A single detected unit. This is a neuron, but because "we cannot
    guarantee that all the spikes assigned to one unit actually originate
    from a single cell", it is called a "unit" rather than a "neuron" """

    def __init__(
            self,
            id: int,
            peak_channel_id: int,
            local_index: int,
            cluster_id: int,
            quality: str,
            firing_rate: float,
            isi_violations: float,
            presence_ratio: float,
            amplitude_cutoff: float,
            mean_waveforms: np.ndarray,
            spike_amplitudes: np.ndarray,
            spike_times: np.ndarray,
            isolation_distance: Optional[float] = None,
            l_ratio: Optional[float] = None,
            d_prime: Optional[float] = None,
            nn_hit_rate: Optional[float] = None,
            nn_miss_rate: Optional[float] = None,
            max_drift: Optional[float] = None,
            cumulative_drift: Optional[float] = None,
            silhouette_score: Optional[float] = None,
            waveform_duration: Optional[float] = None,
            waveform_halfwidth: Optional[float] = None,
            PT_ratio: Optional[float] = None,
            repolarization_slope: Optional[float] = None,
            recovery_slope: Optional[float] = None,
            amplitude: Optional[float] = None,
            spread: Optional[float] = None,
            velocity_above: Optional[float] = None,
            velocity_below: Optional[float] = None,
            snr: Optional[float] = None,
            filter_and_sort_spikes=True
    ):
        super().__init__(name='unit',
                         value=None,
                         is_value_self=True)
        if filter_and_sort_spikes:
            spike_times, spike_amplitudes = _get_filtered_and_sorted_spikes(
                spike_times=spike_times, spike_amplitudes=spike_amplitudes)
        self._id = id
        self._peak_channel_id = peak_channel_id
        self._local_index = local_index
        self._cluster_id = cluster_id
        self._quality = quality
        self._firing_rate = firing_rate
        self._isi_violations = isi_violations
        self._presence_ratio = presence_ratio
        self._amplitude_cutoff = amplitude_cutoff
        self._mean_waveforms = mean_waveforms
        self._spike_amplitudes = spike_amplitudes
        self._spike_times = spike_times
        self._isolation_distance = isolation_distance
        self._l_ratio = l_ratio
        self._d_prime = d_prime
        self._nn_hit_rate = nn_hit_rate
        self._nn_miss_rate = nn_miss_rate
        self._max_drift = max_drift
        self._cumulative_drift = cumulative_drift
        self._silhouette_score = silhouette_score
        self._waveform_duration = waveform_duration
        self._waveform_halfwidth = waveform_halfwidth
        self._PT_ratio = PT_ratio
        self._repolarization_slope = repolarization_slope
        self._recovery_slope = recovery_slope
        self._amplitude = amplitude
        self._spread = spread
        self._velocity_above = velocity_above
        self._velocity_below = velocity_below
        self._snr = snr

    @property
    def id(self) -> int:
        return self._id

    @property
    def peak_channel_id(self) -> int:
        return self._peak_channel_id

    @property
    def local_index(self) -> int:
        return self._local_index

    @property
    def cluster_id(self) -> int:
        return self._cluster_id

    @property
    def quality(self) -> str:
        return self._quality

    @property
    def firing_rate(self) -> float:
        return self._firing_rate

    @property
    def isi_violations(self) -> float:
        """

        Returns
        -------
        Fraction of violating spikes
        """
        return self._isi_violations

    @property
    def presence_ratio(self) -> float:
        """

        Returns
        -------
        Fraction of time unit was present during the session
        """
        return self._presence_ratio

    @property
    def amplitude_cutoff(self) -> float:
        return self._amplitude_cutoff

    @property
    def mean_waveforms(self) -> np.ndarray:
        return self._mean_waveforms

    @property
    def spike_amplitudes(self) -> np.ndarray:
        return self._spike_amplitudes

    @property
    def spike_times(self) -> np.ndarray:
        return self._spike_times

    @property
    def isolation_distance(self) -> Optional[float]:
        return self._isolation_distance

    @property
    def l_ratio(self) -> Optional[float]:
        return self._l_ratio

    @property
    def d_prime(self) -> Optional[float]:
        return self._d_prime

    @property
    def nn_hit_rate(self) -> Optional[float]:
        return self._nn_hit_rate

    @property
    def nn_miss_rate(self) -> Optional[float]:
        return self._nn_miss_rate

    @property
    def max_drift(self) -> Optional[float]:
        return self._max_drift

    @property
    def cumulative_drift(self) -> Optional[float]:
        return self._cumulative_drift

    @property
    def silhouette_score(self) -> Optional[float]:
        return self._silhouette_score

    @property
    def waveform_duration(self) -> Optional[float]:
        return self._waveform_duration

    @property
    def PT_ratio(self) -> Optional[float]:
        return self._PT_ratio

    @property
    def repolarization_slope(self) -> Optional[float]:
        return self._repolarization_slope

    @property
    def recovery_slope(self) -> Optional[float]:
        return self._recovery_slope

    @property
    def amplitude(self) -> Optional[float]:
        return self._amplitude

    @property
    def spread(self) -> Optional[float]:
        return self._spread

    @property
    def velocity_above(self) -> Optional[float]:
        return self._velocity_above

    @property
    def velocity_below(self) -> Optional[float]:
        return self._velocity_below

    @property
    def snr(self) -> Optional[float]:
        return self._snr


def _get_filtered_and_sorted_spikes(
        spike_times: np.ndarray, spike_amplitudes: np.ndarray) -> \
        Tuple[np.ndarray, np.ndarray]:
    """Filter out invalid spike timepoints and sort spike data
    (times + amplitudes) by times.

    Returns
    -------
    A tuple containing filtered and sorted spike times and amplitudes
    """
    valid = spike_times >= 0
    filtered_spike_times = spike_times[valid]
    filtered_spike_amplitudes = spike_amplitudes[valid]

    order = np.argsort(filtered_spike_times)
    sorted_spike_times = filtered_spike_times[order]
    sorted_spike_amplitudes = filtered_spike_amplitudes[order]

    return sorted_spike_times, sorted_spike_amplitudes
