from typing import Dict, Union, List, Optional, Callable
import re
import ast

import pandas as pd
import numpy as np
import xarray as xr
import pynwb

from .ecephys_session_api import EcephysSessionApi
from allensdk.brain_observatory.nwb.nwb_api import NwbApi
from allensdk.brain_observatory.ecephys import get_unit_filter_value
from allensdk.brain_observatory.nwb import check_nwbfile_version
from .._channels import Channels
from ..optotagging import OptotaggingTable
from ..probes import Probes
from ...behavior.data_objects.stimuli.presentations import Presentations

color_triplet_re = re.compile(r"\[(-{0,1}\d*\.\d*,\s*)*(-{0,1}\d*\.\d*)\]")


# TODO: If ecephys write_nwb is revisited, need to re-add `structure_id`
# column and add the structure ids to the nwbfile for the
# add_ecephys_electrodes() function.


class EcephysNwbSessionApi(NwbApi, EcephysSessionApi):

    def __init__(self,
                 path,
                 probe_lfp_paths: Optional[
                     Dict[int, Callable[[], pynwb.NWBFile]]] = None,
                 additional_unit_metrics=None,
                 external_channel_columns=None,
                 **kwargs):

        self.filter_out_of_brain_units = kwargs.pop(
            "filter_out_of_brain_units", True)
        self.filter_by_validity = kwargs.pop("filter_by_validity", True)
        self.amplitude_cutoff_maximum = get_unit_filter_value(
            "amplitude_cutoff_maximum", **kwargs)
        self.presence_ratio_minimum = get_unit_filter_value(
            "presence_ratio_minimum", **kwargs)
        self.isi_violations_maximum = get_unit_filter_value(
            "isi_violations_maximum", **kwargs)

        super(EcephysNwbSessionApi, self).__init__(path, **kwargs)
        self.probe_lfp_paths = probe_lfp_paths

        self.additional_unit_metrics = additional_unit_metrics
        self.external_channel_columns = external_channel_columns

        if hasattr(self, "path") and self.path:
            check_nwbfile_version(
                nwbfile_path=self.path,
                desired_minimum_version="2.2.2",
                warning_msg=(
                    f"It looks like the Visual Coding Neuropixels nwbfile "
                    f"you are trying to access ({self.path})"
                    f"was created by a previous (and incompatible) version of "
                    f"AllenSDK and pynwb. You will need to either 1) use "
                    f"AllenSDK version < 2.0.0 or 2) re-download an updated "
                    f"version of the nwbfile to access the desired data."))

    def test(self):
        """ A minimal test to make sure that this API's NWB file exists and is
        readable. Ecephys NWB files use the required session identifier field
        to store the session id, so this is guaranteed to be present for any
        uncorrupted NWB file.

        Of course, this does not ensure that the file as a whole is correct.
        """
        self.get_ecephys_session_id()

    def get_session_start_time(self):
        return self.nwbfile.session_start_time

    def get_stimulus_presentations(self):
        table = Presentations.from_nwb(nwbfile=self.nwbfile,
                                       add_is_change=False)
        table = table.value

        if "color" in table.columns:
            # .loc breaks on nan values so fill with empty string
            # This is backwards compatible change for older nwb files
            # Newer ones encode nan value here with empty string
            table['color'] = table['color'].fillna('')

            # the color column actually contains two parameters. One is
            # coded as rgb triplets and the other as -1 or 1
            if "color_triplet" not in table.columns:
                table["color_triplet"] = pd.Series("", index=table.index)
            rgb_color_match = table["color"].str.match(color_triplet_re)
            table.loc[rgb_color_match, "color_triplet"] = table.loc[
                rgb_color_match, "color"]
            table.loc[rgb_color_match, "color"] = ""

            # make sure the color column's values are numeric
            table.loc[table["color"] != "", "color"] = table.loc[
                table["color"] != "", "color"].apply(ast.literal_eval)

        return table

    def _probe_nwbfile(self, probe_id: int):
        if self.probe_lfp_paths is None:
            raise TypeError(
                "EcephysNwbSessionApi assumes a split NWB file, with "
                "probewise LFP stored in individual files. "
                "this object was not configured with probe_lfp_paths"
            )
        elif probe_id not in self.probe_lfp_paths:
            raise KeyError(
                f"no probe lfp file path is recorded for probe {probe_id}")

        return self.probe_lfp_paths[probe_id]()

    def get_probes(self) -> pd.DataFrame:
        probes: Union[List, pd.DataFrame] = []
        for k, v in self.nwbfile.electrode_groups.items():
            probes.append({
                'id': v.probe_id,
                'name': v.name,
                'location': v.location,
                "sampling_rate": v.device.sampling_rate,
                "lfp_sampling_rate": v.lfp_sampling_rate,
                "has_lfp_data": v.has_lfp_data
            })
        probes = pd.DataFrame(probes)
        probes = probes.set_index(keys='id', drop=True)
        probes = probes.rename(columns={"name": "description"})
        return probes

    def get_channels(self) -> pd.DataFrame:
        channels = Channels.from_nwb(nwbfile=self.nwbfile)
        channels = channels.to_dataframe(
            external_channel_columns=self.external_channel_columns,
            filter_by_validity=self.filter_by_validity)

        return channels

    def get_mean_waveforms(self) -> Dict[int, np.ndarray]:
        probes = Probes.from_nwb(nwbfile=self.nwbfile)
        return probes.mean_waveforms

    def get_spike_times(self) -> Dict[int, np.ndarray]:
        probes = Probes.from_nwb(nwbfile=self.nwbfile)
        return probes.spike_times

    def get_spike_amplitudes(self) -> Dict[int, np.ndarray]:
        probes = Probes.from_nwb(nwbfile=self.nwbfile)
        return probes.spike_amplitudes

    def get_units(self) -> pd.DataFrame:
        probes = Probes.from_nwb(nwbfile=self.nwbfile)
        return probes.get_units_table(
            filter_by_validity=self.filter_by_validity,
            filter_out_of_brain_units=self.filter_out_of_brain_units,
            amplitude_cutoff_maximum=self.amplitude_cutoff_maximum,
            presence_ratio_minimum=self.presence_ratio_minimum,
            isi_violations_maximum=self.isi_violations_maximum
        )

    def get_lfp(self, probe_id: int) -> xr.DataArray:
        lfp_file = self._probe_nwbfile(probe_id)
        lfp = lfp_file.get_acquisition(f'probe_{probe_id}_lfp')
        series = lfp.get_electrical_series(f'probe_{probe_id}_lfp_data')

        electrodes = lfp_file.electrodes.to_dataframe()

        data = series.data[:]
        timestamps = series.timestamps[:]

        return xr.DataArray(
            name="LFP",
            data=data,
            dims=['time', 'channel'],
            coords=[timestamps, electrodes.index.values]
        )

    def get_running_speed(self, include_rotation=False) -> pd.DataFrame:
        running_module = self.nwbfile.get_processing_module("running")
        running_speed_series = running_module["running_speed"]
        running_speed_start_times = running_speed_series.timestamps[:]

        running_speed_end_series = running_module["running_speed_end_times"]
        running_speed_end_times = running_speed_end_series.timestamps[:]

        running = pd.DataFrame({
            "start_time": running_speed_start_times,
            "end_time": running_speed_end_times,
            "velocity": running_speed_series.data[:]
        })

        if include_rotation:
            rotation_series = running_module["running_wheel_rotation"]
            running["net_rotation"] = rotation_series.data[:]

        return running

    def get_raw_running_data(self):
        rotation_series = self.nwbfile.get_acquisition(
            "raw_running_wheel_rotation")
        signal_voltage_series = self.nwbfile.get_acquisition(
            "running_wheel_signal_voltage")
        supply_voltage_series = self.nwbfile.get_acquisition(
            "running_wheel_supply_voltage")

        return pd.DataFrame({
            "frame_time": rotation_series.timestamps[:],
            "net_rotation": rotation_series.data[:],
            "signal_voltage": signal_voltage_series.data[:],
            "supply_voltage": supply_voltage_series.data[:]
        })

    def get_rig_metadata(self) -> Optional[dict]:
        try:
            et_mod = self.nwbfile.get_processing_module(
                "eye_tracking_rig_metadata")
        except KeyError as e:
            print(
                f"This ecephys session '{int(self.nwbfile.identifier)}' has "
                f"no eye tracking rig metadata. (NWB error: {e})")
            return None

        meta = et_mod.get_data_interface("eye_tracking_rig_metadata")

        rig_geometry = pd.DataFrame({
            f"monitor_position_{meta.monitor_position__unit}":
                meta.monitor_position,
            f"camera_position_{meta.camera_position__unit}":
                meta.camera_position,
            f"led_position_{meta.led_position__unit}": meta.led_position,
            f"monitor_rotation_{meta.monitor_rotation__unit}":
                meta.monitor_rotation,
            f"camera_rotation_{meta.camera_rotation__unit}":
                meta.camera_rotation
        })

        rig_geometry = rig_geometry.rename(index={0: 'x', 1: 'y', 2: 'z'})

        returned_metadata = {
            "geometry": rig_geometry,
            "equipment": meta.equipment
        }

        return returned_metadata

    def get_screen_gaze_data(self, include_filtered_data=False) -> \
            Optional[pd.DataFrame]:
        try:
            rgm_mod = self.nwbfile.get_processing_module("raw_gaze_mapping")
            fgm_mod = self.nwbfile.get_processing_module(
                "filtered_gaze_mapping")
        except KeyError as e:
            print(
                f"This ecephys session '{int(self.nwbfile.identifier)}' has "
                f"no eye tracking data. (NWB error: {e})")
            return None

        raw_eye_area_ts = rgm_mod.get_data_interface("eye_area")
        raw_pupil_area_ts = rgm_mod.get_data_interface("pupil_area")
        raw_screen_coordinates_ts = rgm_mod.get_data_interface(
            "screen_coordinates")
        raw_screen_coordinates_spherical_ts = rgm_mod.get_data_interface(
            "screen_coordinates_spherical")

        filtered_eye_area_ts = fgm_mod.get_data_interface("eye_area")
        filtered_pupil_area_ts = fgm_mod.get_data_interface("pupil_area")
        filtered_screen_coordinates_ts = fgm_mod.get_data_interface(
            "screen_coordinates")
        filtered_screen_coordinates_spherical_ts = fgm_mod.get_data_interface(
            "screen_coordinates_spherical")

        gaze_data = {
            "raw_eye_area": raw_eye_area_ts.data[:],
            "raw_pupil_area": raw_pupil_area_ts.data[:],
            "raw_screen_coordinates_x_cm":
                raw_screen_coordinates_ts.data[:, 1],
            "raw_screen_coordinates_y_cm":
                raw_screen_coordinates_ts.data[:, 0],
            "raw_screen_coordinates_spherical_x_deg":
                raw_screen_coordinates_spherical_ts.data[:, 1],
            "raw_screen_coordinates_spherical_y_deg":
                raw_screen_coordinates_spherical_ts.data[:, 0]
        }

        if include_filtered_data:
            gaze_data.update(
                {
                    "filtered_eye_area": filtered_eye_area_ts.data[:],
                    "filtered_pupil_area": filtered_pupil_area_ts.data[:],
                    "filtered_screen_coordinates_x_cm":
                        filtered_screen_coordinates_ts.data[
                                                        :, 1],
                    "filtered_screen_coordinates_y_cm":
                        filtered_screen_coordinates_ts.data[
                                                        :, 0],
                    "filtered_screen_coordinates_spherical_x_deg":
                        filtered_screen_coordinates_spherical_ts.data[
                                                                   :, 1],
                    "filtered_screen_coordinates_spherical_y_deg":
                        filtered_screen_coordinates_spherical_ts.data[
                                                                   :, 0]
                }
            )

        index = pd.Index(data=raw_eye_area_ts.timestamps[:], name="Time (s)")
        return pd.DataFrame(gaze_data, index=index)

    def get_pupil_data(self) -> Optional[pd.DataFrame]:
        try:
            et_mod = self.nwbfile.get_processing_module("eye_tracking")
            rgm_mod = self.nwbfile.get_processing_module("raw_gaze_mapping")
        except KeyError as e:
            print(
                f"This ecephys session '{int(self.nwbfile.identifier)}' has "
                f"no eye tracking data. (NWB error: {e})")
            return None

        cr_ellipse_fits = et_mod.get_data_interface(
            "cr_ellipse_fits").to_dataframe()
        eye_ellipse_fits = et_mod.get_data_interface(
            "eye_ellipse_fits").to_dataframe()
        pupil_ellipse_fits = et_mod.get_data_interface(
            "pupil_ellipse_fits").to_dataframe()

        # NOTE: ellipse fit "height" and "width" parameters describe the
        # "half-height" and "half-width" of fitted ellipse.
        eye_tracking_data = {
            "corneal_reflection_center_x": cr_ellipse_fits["center_x"].values,
            "corneal_reflection_center_y": cr_ellipse_fits["center_y"].values,
            "corneal_reflection_height": 2 * cr_ellipse_fits["height"].values,
            "corneal_reflection_width": 2 * cr_ellipse_fits["width"].values,
            "corneal_reflection_phi": cr_ellipse_fits["phi"].values,

            "pupil_center_x": pupil_ellipse_fits["center_x"].values,
            "pupil_center_y": pupil_ellipse_fits["center_y"].values,
            "pupil_height": 2 * pupil_ellipse_fits["height"].values,
            "pupil_width": 2 * pupil_ellipse_fits["width"].values,
            "pupil_phi": pupil_ellipse_fits["phi"].values,

            "eye_center_x": eye_ellipse_fits["center_x"].values,
            "eye_center_y": eye_ellipse_fits["center_y"].values,
            "eye_height": 2 * eye_ellipse_fits["height"].values,
            "eye_width": 2 * eye_ellipse_fits["width"].values,
            "eye_phi": eye_ellipse_fits["phi"].values
        }

        timestamps = rgm_mod.get_data_interface("eye_area").timestamps[:]
        index = pd.Index(data=timestamps, name="Time (s)")
        return pd.DataFrame(eye_tracking_data, index=index)

    def get_ecephys_session_id(self) -> int:
        return int(self.nwbfile.identifier)

    def get_current_source_density(self, probe_id):
        csd_mod = self._probe_nwbfile(probe_id).get_processing_module(
            "current_source_density")
        nwb_csd = csd_mod["ecephys_csd"]
        csd_data = nwb_csd.time_series.data[
                   :].T  # csd data stored as (timepoints x channels) but we
        # want (channels x timepoints)

        csd = xr.DataArray(
            name="CSD",
            data=csd_data,
            dims=["virtual_channel_index", "time"],
            coords={
                "virtual_channel_index": np.arange(csd_data.shape[0]),
                "time": nwb_csd.time_series.timestamps[:],
                "vertical_position": (("virtual_channel_index",),
                                      nwb_csd.virtual_electrode_y_positions),
                "horizontal_position": (("virtual_channel_index",),
                                        nwb_csd.virtual_electrode_x_positions)
            }
        )
        return csd

    def get_optogenetic_stimulation(self) -> pd.DataFrame:
        table = OptotaggingTable.from_nwb(nwbfile=self.nwbfile)
        return table.value

    def get_metadata(self):
        nwb_subject = self.nwbfile.subject
        metadata = {
            "specimen_name": nwb_subject.specimen_name,
            "age_in_days": nwb_subject.age_in_days,
            "full_genotype": nwb_subject.genotype,
            "strain": nwb_subject.strain,
            "sex": nwb_subject.sex,
            "stimulus_name": self.nwbfile.stimulus_notes,
            "subject_id": nwb_subject.subject_id,
            "age": nwb_subject.age,
            "species": nwb_subject.species
        }
        return metadata
